package com.cbr.hdt;

import com.cbr.hdt.impl.Triple;
import com.cbr.hdt.impl.DictEntry;
import com.cbr.hdt.impl.spark.SparkDictSection;
import com.hadoop.mapreduce.LzoTextInputFormat;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.lib.input.TextInputFormat;
import org.apache.spark.SparkConf;
import org.apache.spark.api.java.JavaPairRDD;
import org.apache.spark.api.java.JavaRDD;
import org.apache.spark.api.java.JavaSparkContext;
import org.apache.spark.api.java.function.Function;
import org.apache.spark.api.java.function.Function2;
import org.apache.spark.api.java.function.PairFlatMapFunction;
import org.apache.spark.api.java.function.PairFunction;
import org.apache.spark.broadcast.Broadcast;
import org.apache.spark.sql.Dataset;
import org.apache.spark.sql.Row;
import org.apache.spark.sql.SQLContext;
import org.apache.spark.sql.SaveMode;
import org.apache.spark.sql.types.StructType;
import com.cbr.hdt.impl.spark.SparkBitMapTriples;
import com.cbr.hdt.impl.spark.SparkFourSectionDictionary;
import com.cbr.hdt.spark.MutablePrimitive;
import com.cbr.hdt.spark.Stats;
import com.cbr.hdt.spark.StatsAccumulator;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import org.apache.spark.sql.RowFactory;
import org.apache.spark.storage.StorageLevel;
import org.rdfhdt.hdt.hdt.HDT;
import org.rdfhdt.hdt.hdt.HDTManager;
import org.rdfhdt.hdt.hdt.HDTVocabulary;
import org.rdfhdt.hdt.header.Header;
import org.rdfhdt.hdt.header.HeaderFactory;
import org.rdfhdt.hdt.options.ControlInfo;
import org.rdfhdt.hdt.options.ControlInformation;
import org.rdfhdt.hdt.options.HDTSpecification;
import org.rdfhdt.hdt.triples.IteratorTripleString;
import org.rdfhdt.hdt.triples.TripleString;
import scala.Tuple2;

/**
 *
 * @author Carlos Barrales Ruiz
 * <a href="mailto:cbruiz@gmail.com">cbruiz@gmail.com</a>
 */
public class HDTSpark {

    /**
     *
     * @param envFlag env flag
     * @param defaultValue default value
     * @return result
     */
    public static String getEnvParam(final String envFlag, final String defaultValue) {
        final String envFlagValue = System.getenv(envFlag);
        if (envFlagValue == null) {
            return defaultValue;
        } else {
            return envFlagValue;
        }
    }

    /**
     * Entry points
     *
     * @param args arguments
     * @throws java.lang.Exception
     */
    public static void main(String[] args) throws Exception {

        final boolean debugMode = "1".equals(getEnvParam("DEBUG", "0"));
        final boolean generateDict = "1".equals(getEnvParam("GENERATE_DICT", "1"));
        final boolean encodeTriples = "1".equals(getEnvParam("ENCODE_TRIPLES", "1"));
        final boolean generateHDT = "1".equals(getEnvParam("GENERATE_HDT", "1"));
        final boolean testHDT = "1".equals(getEnvParam("TEST_HDT", "0"));
        final String outputFormat = getEnvParam("FORMAT", "parquet");
        final boolean doCache = "1".equals(getEnvParam("CACHE", "0"));
        final boolean doPersist = "1".equals(getEnvParam("PERSIST", "0"));
        final boolean doRepartition = "1".equals(getEnvParam("REPARTITION", "1"));
        final boolean cacheInput = "1".equals(getEnvParam("CACHE_INPUT", "0"));
        int numPartitions = Integer.parseInt(getEnvParam("NUM_PARTITIONS", "-1"));
        final int outputParts = Integer.parseInt(getEnvParam("NUM_OUTPUT_PARTS", "-1"));

        final Lock lock = new ReentrantLock();
        final Condition backgroundOperation = lock.newCondition();
        final MutablePrimitive<Boolean> dictionaryReady = new MutablePrimitive<>(false);
        final MutablePrimitive<Boolean> triplesReady = new MutablePrimitive<>(false);

        final JoinMethod joinMethod = JoinMethod.valueOf(getEnvParam("JOIN_METHOD", "JOIN_DATAFRAMES"));

        boolean standAlone = false;
        String inputPath = null;
        String outPath = null;

        for (String arg : args) {
            if (arg.equalsIgnoreCase("--standalone")) {
                standAlone = true;
            } else if (!arg.startsWith("--")) {
                if (inputPath == null) {
                    inputPath = arg;
                } else if (outPath == null) {
                    outPath = arg;
                }
            }
        }
        if (inputPath == null || outPath == null) {
            System.err.println("Uso: java -jar <jar> [--standalone] <inputFile> <outputFile>");
            System.err.println("\tExample: java hdt-spark.jar hdfs://user/user1/input.dsv hdfs://user/user1/output.dsv");
            System.exit(1);
        }

        System.out.println(String.format("HDT-SPARK processing %s", inputPath));

        final String outputPath = outPath.endsWith("/") ? outPath : (outPath + "/");

        final SparkConf conf = new SparkConf().setAppName("HDT-SPARK");
        if (numPartitions > 0) {
            conf.set("spark.sql.shuffle.partitions", Integer.toString(numPartitions));
            conf.set("spark.default.parallelism", Integer.toString(numPartitions));
            conf.set("spark.sql.shuffle.partitions", Integer.toString(numPartitions));
        }
        if (standAlone) {
            conf.setMaster("local");
        }

        final JavaSparkContext jsc = new JavaSparkContext(conf);
        final SQLContext sqlContext = new SQLContext(jsc);

        final Map<String, Stats> datasetStats = new HashMap<>();

        final StatsAccumulator tripleStats = new StatsAccumulator();

        final StatsAccumulator subjStats = new StatsAccumulator();
        final StatsAccumulator predStats = new StatsAccumulator();
        final StatsAccumulator objStats = new StatsAccumulator();
        final StatsAccumulator soStats = new StatsAccumulator();

        final StatsAccumulator subjKeyStats = new StatsAccumulator();
        final StatsAccumulator predKeyStats = new StatsAccumulator();
        final StatsAccumulator objKeyStats = new StatsAccumulator();
        final StatsAccumulator soKeyStats = new StatsAccumulator();

        jsc.sc().register(tripleStats, "tripleStats");

        jsc.sc().register(subjStats, "subjStats");
        jsc.sc().register(predStats, "predStats");
        jsc.sc().register(objStats, "objStats");
        jsc.sc().register(soStats, "soStats");

        jsc.sc().register(subjKeyStats, "subjKeyStats");
        jsc.sc().register(predKeyStats, "predKeyStats");
        jsc.sc().register(objKeyStats, "objKeyStats");
        jsc.sc().register(soKeyStats, "soKeyStats");

        // END Spark kernel
        //===================================
        //===================================
        // BEGIN Spark kernel
        Configuration jobConf = jsc.hadoopConfiguration();
        jobConf.set("io.compression.codecs",
                "org.apache.hadoop.io.compress.DefaultCodec,"
                + "org.apache.hadoop.io.compress.GzipCodec,"
                + "org.apache.hadoop.io.compress.BZip2Codec,"
                + "com.hadoop.compression.lzo.LzoCodec,"
                + "com.hadoop.compression.lzo.LzopCodec,"
                + "org.apache.hadoop.io.compress.SnappyCodec");
        jobConf.set("io.compression.codec.lzo.class", "com.hadoop.compression.lzo.LzopCodec");

        int numExecutors = jsc.getConf().getInt("spark.executor.instances", 10);
        int numCores = jsc.getConf().getInt("spark.executor.cores", 4);
        int replicationFactor = jsc.getConf().getInt("spark.hadoop.dfs.replication", 1);

        JavaPairRDD<LongWritable, Text> input = jsc.newAPIHadoopFile(inputPath, inputPath.endsWith(".lzo") ? LzoTextInputFormat.class : TextInputFormat.class, LongWritable.class, Text.class, jobConf);

        if (numPartitions == -1) {
            long numFiles = input.count();
            numPartitions = (int) Math.min(500, Math.max(numFiles / 4451696, 1));
            System.out.println(
                    String.format("Job distribution details:\n\tNumExecutors = %d\n\tNumCores = %d\n\tReplicationFactor = %d\n\t---\n\tNumTuples = %d\n\tNumPartitions = %d",
                            numExecutors, numCores, replicationFactor, numFiles, numPartitions
                    ));
        } else if (doRepartition) {
            System.out.println(
                    String.format("Job distribution details:\n\tNumExecutors = %d\n\tNumCores = %d\n\tReplicationFactor = %d\n\t---\n\tNumTuples = N/A\n\tNumPartitions = %d",
                            numExecutors, numCores, replicationFactor, numPartitions
                    ));
        } else {
            System.out.println(
                    String.format("Job distribution details:\n\tNumExecutors = %d\n\tNumCores = %d\n\tReplicationFactor = %d\n\t---\n\tNumTuples = N/A\n\tNumPartitions = N/A",
                            numExecutors, numCores, replicationFactor, numPartitions
                    ));
        }
        JavaRDD<Triple> tripleRDD;
        if (joinMethod == JoinMethod.COGROUP) {
            tripleRDD = input.values().zipWithIndex().map(new Function<Tuple2<Text, Long>, Triple>() {
                @Override
                public Triple call(Tuple2<Text, Long> t) throws Exception {
                    TripleString ts = new TripleString();
                    ts.read(t._1.toString());
                    Triple result = new Triple(ts.getSubject().toString(), ts.getPredicate().toString(), ts.getObject().toString());
                    result.setIndex(t._2);
                    return result;
                }
            });
        } else {
            tripleRDD = input.values().map(new Function<Text, Triple>() {
                @Override
                public Triple call(Text t) throws Exception {
                    TripleString ts = new TripleString();
                    ts.read(t.toString());
                    Triple result = new Triple(ts.getSubject().toString(), ts.getPredicate().toString(), ts.getObject().toString());
                    return result;
                }
            });
        }

        if (doRepartition && numPartitions > 1) {
            tripleRDD = tripleRDD.coalesce(numPartitions, false);
        }
        if (cacheInput) {
            if (doCache) {
                tripleRDD = tripleRDD.cache();
            } else if (doPersist) {
                tripleRDD = tripleRDD.persist(StorageLevel.MEMORY_AND_DISK_2());
            }
        }
        final JavaRDD<Triple> fTripleRDD = tripleRDD;
        final int nparts = numPartitions;
        final Map<String, Dataset<Row>> dict = new HashMap<>();
        new Thread() {
            @Override
            public void run() {
                if (generateDict) {
                    try {
                        generateDictionary(outputParts, fTripleRDD, doRepartition, nparts, doCache, doPersist, soStats, soKeyStats, subjStats, subjKeyStats, predStats, predKeyStats, objStats, objKeyStats, sqlContext, outputFormat, outputPath, datasetStats);
                    } catch (Exception ex) {
                        ex.printStackTrace(System.err);
                        System.exit(1);
                    }
                }

                System.out.println("Loading dictionary...");

                StructType schema = new StructType()
                        .add("key", "STRING", true)
                        .add("value", "LONG", true);

                dict.put("so", sqlContext.read().format(outputFormat).schema(schema).load(outputPath + "so.dict"));
                dict.put("su", sqlContext.read().format(outputFormat).schema(schema).load(outputPath + "su.dict"));
                dict.put("ob", sqlContext.read().format(outputFormat).schema(schema).load(outputPath + "ob.dict"));
                dict.put("pr", sqlContext.read().format(outputFormat).schema(schema).load(outputPath + "pr.dict"));

                if (debugMode) {
                    System.out.println("SO:");
                    dict.get("so").show(10);
                    System.out.println("SU:");
                    dict.get("su").show(10);
                    System.out.println("PR:");
                    dict.get("pr").show(10);
                    System.out.println("OB:");
                    dict.get("ob").show(10);
                }
                System.out.println("Dictionary loaded");
                lock.lock();
                try {
                    dictionaryReady.set(true);
                    backgroundOperation.signalAll();
                } finally {
                    lock.unlock();
                }
            }

        }.start();

        if (encodeTriples) {
            new Thread() {
                @Override
                public void run() {
                    System.out.println("Encoding triples [Waiting for dictionary job]...");
                    lock.lock();
                    try {
                        while (!dictionaryReady.get()) {
                            backgroundOperation.await();
                        }
                    } catch (Exception ex) {
                        ex.printStackTrace(System.err);
                        System.exit(1);
                    } finally {
                        lock.unlock();
                    }

                    System.out.println("Encoding triples...");
                    encodeTriples(outputParts, joinMethod, sqlContext, fTripleRDD, datasetStats, dict, tripleStats, outputFormat, outputPath, jsc);

                    System.out.println("Triples encoded");
                    lock.lock();
                    try {
                        triplesReady.set(true);
                        backgroundOperation.signalAll();
                    } finally {
                        lock.unlock();
                    }

                }
            }.start();
        }
        else {
            lock.lock();
            try {
                triplesReady.set(true);
                backgroundOperation.signalAll();
            } finally {
                lock.unlock();
            }
        }

        if (generateHDT) {
            long tHeader = 0;
            long tDict = 0;
            long tTriples = 0;

            long t1 = System.currentTimeMillis();

            StructType schema = new StructType()
                    .add("sCod", "LONG", true)
                    .add("pCod", "LONG", true)
                    .add("oCod", "LONG", true)
                    .add("rn", "LONG", true);

            HDTSpecification spec = new HDTSpecification();

            { // Write HDT
                Header header = HeaderFactory.createHeader(spec);

                OutputStream output = new FileOutputStream("output.hdt"); // write locally
                ControlInfo ci = new ControlInformation();

                ci.clear();
                ci.setType(ControlInfo.Type.GLOBAL);
                ci.setFormat(HDTVocabulary.HDT_CONTAINER);
                ci.save(output);

                tHeader = System.currentTimeMillis() - t1;
                System.out.println("Generating HDT [Waiting for dictionary job]...");
                lock.lock();
                try {
                    while (!dictionaryReady.get()) {
                        backgroundOperation.await();
                    }
                } finally {
                    lock.unlock();
                }
                t1 = System.currentTimeMillis();
                Dataset<Row> so = dict.get("so").select("key", "value");
                Dataset<Row> su = dict.get("su").select("key", "value");
                Dataset<Row> pr = dict.get("pr").select("key", "value");
                Dataset<Row> ob = dict.get("ob").select("key", "value");
                if (debugMode) {
                    System.out.println("SO:");
                    so.orderBy(so.col("value").asc()).show(10);
                    System.out.println("SU:");
                    su.orderBy(su.col("value").asc()).show(10);
                    System.out.println("PR:");
                    pr.orderBy(pr.col("value").asc()).show(10);
                    System.out.println("OB:");
                    ob.orderBy(ob.col("value").asc()).show(10);
                }

                SparkDictSection s = new SparkDictSection(spec, "Subject", jsc.hadoopConfiguration(), outputPath + "su.dict", su);
                SparkDictSection p = new SparkDictSection(spec, "Predicate", jsc.hadoopConfiguration(), outputPath + "pr.dict", pr);
                SparkDictSection o = new SparkDictSection(spec, "Object", jsc.hadoopConfiguration(), outputPath + "ob.dict", ob);
                SparkDictSection sh = new SparkDictSection(spec, "Shared", jsc.hadoopConfiguration(), outputPath + "so.dict", so);

                SparkFourSectionDictionary dictionary = new SparkFourSectionDictionary(spec, s, p, o, sh);
                dictionary.loadFromSpark(null, datasetStats);
                dictionary.populateHeader(header, "_:dictionary");

                System.out.println(String.format("\tProcessing Header"));
                ci.clear();
                ci.setType(ControlInfo.Type.HEADER);
                header.save(output, ci, null);

                System.out.println(String.format("\tProcessing Dictionary"));
                ci.clear();
                ci.setType(ControlInfo.Type.DICTIONARY);
                
                dictionary.save(output, ci, null);

                tDict = System.currentTimeMillis() - t1;
                System.out.println("Generating HDT [Waiting for triples job]...");
                lock.lock();
                try {
                    while (!triplesReady.get()) {
                        backgroundOperation.await();
                    }
                } finally {
                    lock.unlock();
                }
                t1 = System.currentTimeMillis();
                System.out.println(String.format("\tProcessing Triples"));
                Dataset<Row> dfEncTriples = sqlContext.read().format(outputFormat).schema(schema).load(outputPath + "triples.dat");
                SparkBitMapTriples triples = new SparkBitMapTriples(spec, jsc.hadoopConfiguration(), outputPath + "triples.dat", dfEncTriples);
                ci.clear();
                ci.setType(ControlInfo.Type.TRIPLES);
                triples.load(null, datasetStats);
                triples.save(output, ci, null);
                tTriples = System.currentTimeMillis() - t1;

                System.out.println(String.format("\tHDT serialized in %.03f secs (%.03f + %.03f + %.03f).",
                        ((double) (tHeader + tDict + tTriples)) / 1000.0,
                        ((double) (tHeader)) / 1000.0,
                        ((double) (tDict)) / 1000.0,
                        ((double) (tTriples)) / 1000.0
                ));
            }
        }
        if (testHDT) {
            System.out.println("Testing HDT...");

            long t1 = System.currentTimeMillis();

            HDT hdt = HDTManager.loadHDT("output.hdt", null);
            IteratorTripleString it = hdt.search(null, null, null);
            int num = 0;
            System.out.println("\tSample dump:");
            while (it.hasNext()) {
                TripleString st = it.next();
                System.err.println(String.format("\t%s", st));
                if (num++ == 10) {
                    break;
                }
            }
            long tdiff = System.currentTimeMillis() - t1;
            System.out.println(String.format("\tHDT tested in %.03f secs.", (tdiff / 1000.0f)));
        }
        // END Spark kernel
        //===================================

        System.out.println("Stopping context");
        jsc.stop();
        System.out.println("Completed");
        System.exit(0);
    }

    /**
     *
     * @param outputParts
     * @param joinMethod
     * @param sqlContext
     * @param tripleRDD
     * @param datasetStats
     * @param dict
     * @param tripleStats
     * @param outputFormat
     * @param outputPath
     * @param jsc
     */
    private static void encodeTriples(final int outputParts, final JoinMethod joinMethod, final SQLContext sqlContext, JavaRDD<Triple> tripleRDD, Map<String, Stats> datasetStats, Map<String, Dataset<Row>> dict, final StatsAccumulator tripleStats, final String outputFormat, final String outputPath, final JavaSparkContext jsc) {
        if (joinMethod == JoinMethod.JOIN_DATAFRAMES) {

            System.out.println("Encoding triples with JOIN_DATAFRAMES...");

            long t1 = System.currentTimeMillis();
            Dataset<Row> tripleDF = sqlContext.createDataFrame(tripleRDD, Triple.class);

            Long soOffset = 1L;
            if (datasetStats.containsKey("soStats")) {
                soOffset += datasetStats.get("soStats").getCount();
            } else {
                soOffset += dict.get("so").count();
            }

            Dataset<Row> so = dict.get("so").withColumn("value2", dict.get("so").col("value").plus(1)).drop("value").withColumnRenamed("value2", "value");
            Dataset<Row> su = dict.get("su").withColumn("value2", dict.get("su").col("value").plus(soOffset)).drop("value").withColumnRenamed("value2", "value");
            Dataset<Row> pr = dict.get("pr").withColumn("value2", dict.get("pr").col("value").plus(1)).drop("value").withColumnRenamed("value2", "value");
            Dataset<Row> ob = dict.get("ob").withColumn("value2", dict.get("ob").col("value").plus(soOffset)).drop("value").withColumnRenamed("value2", "value");

            su = su.union(so);
            ob = ob.union(so);

            Dataset<Row> df1 = tripleDF.drop("sCod").join(su, tripleDF.col("subject").equalTo(su.col("key"))).withColumnRenamed("value", "sCod").drop("subject").drop("key").drop("value");
            Dataset<Row> df2 = df1.drop("oCod").join(ob, tripleDF.col("object").equalTo(ob.col("key"))).withColumnRenamed("value", "oCod").drop("object").drop("key").drop("value");
            Dataset<Row> df3 = df2.drop("pCod").join(pr, tripleDF.col("predicate").equalTo(pr.col("key"))).withColumnRenamed("value", "pCod").drop("predicate").drop("key").drop("value");

            Dataset<Row> dfResult = df3;

            StructType schema = new StructType()
                    .add("sCod", "LONG", false)
                    .add("pCod", "LONG", false)
                    .add("oCod", "LONG", false)
                    .add("rn", "LONG", false);

            JavaRDD<Row> rdd = dfResult.select("sCod", "pCod", "oCod").orderBy(dfResult.col("sCod").asc(), dfResult.col("pCod").asc(), dfResult.col("oCod").asc())
                    .toJavaRDD().zipWithIndex().map(new Function<Tuple2<Row, Long>, Row>() {
                        @Override
                        public Row call(Tuple2<Row, Long> t1) throws Exception {
                            tripleStats.add(new Stats(1));
                            return RowFactory.create(t1._1.getLong(0), t1._1.getLong(1), t1._1.getLong(2), t1._2);
                        }
                    });

            dfResult = sqlContext.createDataFrame(rdd.rdd(), schema);
            if (outputParts > 0) {
                dfResult = dfResult.coalesce(outputParts).orderBy(dfResult.col("rn").asc());
            }
            dfResult.select("sCod", "pCod", "oCod").write().mode(SaveMode.Overwrite).format(outputFormat).save(outputPath + "triples.dat");

            datasetStats.put("tripleStats", tripleStats.value());
            long tdiff = System.currentTimeMillis() - t1;
            System.out.println(String.format("\t%d triples encoded in %.03f secs.", tripleStats.value().getCount(), (tdiff / 1000.0f)));

        } else if (joinMethod == JoinMethod.COGROUP) {

            System.out.println("Encoding triples with COGROUP...");

            long t1 = System.currentTimeMillis();

            Long soOffset = 1L;
            if (datasetStats.containsKey("soStats")) {
                soOffset += datasetStats.get("soStats").getCount();
            } else {
                soOffset += dict.get("so").count();
            }

            Dataset<Row> so = dict.get("so").withColumn("value2", dict.get("so").col("value").plus(1)).drop("value").withColumnRenamed("value2", "value");
            Dataset<Row> su = dict.get("su").withColumn("value2", dict.get("su").col("value").plus(soOffset)).drop("value").withColumnRenamed("value2", "value");
            Dataset<Row> pr = dict.get("pr").withColumn("value2", dict.get("pr").col("value").plus(1)).drop("value").withColumnRenamed("value2", "value");
            Dataset<Row> ob = dict.get("ob").withColumn("value2", dict.get("ob").col("value").plus(soOffset)).drop("value").withColumnRenamed("value2", "value");

            su = su.union(so);
            ob = ob.union(so);

            JavaPairRDD<String, Long> suRDD = su.javaRDD().mapToPair(new PairFunction<Row, String, Long>() {
                @Override
                public Tuple2<String, Long> call(Row t) throws Exception {
                    return new Tuple2<>(t.getString(0), t.getLong(1));
                }
            });

            JavaPairRDD<String, Long> prRDD = pr.javaRDD().mapToPair(new PairFunction<Row, String, Long>() {
                @Override
                public Tuple2<String, Long> call(Row t) throws Exception {
                    return new Tuple2<>(t.getString(0), t.getLong(1));
                }
            });

            JavaPairRDD<String, Long> obRDD = ob.javaRDD().mapToPair(new PairFunction<Row, String, Long>() {
                @Override
                public Tuple2<String, Long> call(Row t) throws Exception {
                    return new Tuple2<>(t.getString(0), t.getLong(1));
                }
            });
            /// ==========

            JavaPairRDD<String, Long> rs = tripleRDD.mapToPair(new PairFunction<Triple, String, Long>() {
                @Override
                public Tuple2<String, Long> call(Triple t) throws Exception {
                    return new Tuple2<>(t.getSubject(), t.getIndex());
                }
            });

            JavaPairRDD<String, Long> rp = tripleRDD.mapToPair(new PairFunction<Triple, String, Long>() {
                @Override
                public Tuple2<String, Long> call(Triple t) throws Exception {
                    return new Tuple2<>(t.getPredicate(), t.getIndex());
                }
            });

            JavaPairRDD<String, Long> ro = tripleRDD.mapToPair(new PairFunction<Triple, String, Long>() {
                @Override
                public Tuple2<String, Long> call(Triple t) throws Exception {
                    return new Tuple2<>(t.getObject(), t.getIndex());
                }
            });
            /// ==========

            JavaPairRDD<String, Tuple2<Iterable<Long>, Iterable<Long>>> senc = suRDD.cogroup(rs);
            JavaPairRDD<String, Tuple2<Iterable<Long>, Iterable<Long>>> penc = prRDD.cogroup(rp);
            JavaPairRDD<String, Tuple2<Iterable<Long>, Iterable<Long>>> oenc = obRDD.cogroup(ro);

            /// ============= Column encoding
            JavaPairRDD<Long, Long> senc2 = senc.values().flatMapToPair(
                    new PairFlatMapFunction<Tuple2<Iterable<Long>, Iterable<Long>>, Long, Long>() {
                @Override
                public Iterator<Tuple2<Long, Long>> call(final Tuple2<Iterable<Long>, Iterable<Long>> tuple) throws Exception {
                    ArrayList<Tuple2<Long, Long>> t = new ArrayList<>();
                    for (Long code : tuple._1) {
                        for (Long rowId : tuple._2) {
                            t.add(new Tuple2<>(rowId, code));
                        }
                    }
                    return t.iterator();
                }
            });
            JavaPairRDD<Long, Long> penc2 = penc.values().flatMapToPair(
                    new PairFlatMapFunction<Tuple2<Iterable<Long>, Iterable<Long>>, Long, Long>() {
                @Override
                public Iterator<Tuple2<Long, Long>> call(final Tuple2<Iterable<Long>, Iterable<Long>> tuple) throws Exception {
                    ArrayList<Tuple2<Long, Long>> t = new ArrayList<>();
                    for (Long code : tuple._1) {
                        for (Long rowId : tuple._2) {
                            t.add(new Tuple2<>(rowId, code));
                        }
                    }
                    return t.iterator();
                }
            });
            JavaPairRDD<Long, Long> oenc2 = oenc.values().flatMapToPair(
                    new PairFlatMapFunction<Tuple2<Iterable<Long>, Iterable<Long>>, Long, Long>() {
                @Override
                public Iterator<Tuple2<Long, Long>> call(final Tuple2<Iterable<Long>, Iterable<Long>> tuple) throws Exception {
                    ArrayList<Tuple2<Long, Long>> t = new ArrayList<>();
                    for (Long code : tuple._1) {
                        for (Long rowId : tuple._2) {
                            t.add(new Tuple2<>(rowId, code));
                        }
                    }
                    return t.iterator();
                }
            });

            JavaRDD<Triple> encodedRDD = senc2.cogroup(penc2).cogroup(oenc2).map(new Function< Tuple2<Long, Tuple2<Iterable<Tuple2<Iterable<Long>, Iterable<Long>>>, Iterable<Long>>>, Triple>() {
                @Override
                public Triple call(Tuple2<Long, Tuple2<Iterable<Tuple2<Iterable<Long>, Iterable<Long>>>, Iterable<Long>>> tuple) throws Exception {
                    Triple triple = new Triple();
                    triple.setIndex(tuple._1);
                    Tuple2<Iterable<Long>, Iterable<Long>> spIt = tuple._2._1.iterator().next();
                    triple.setsCod(spIt._1.iterator().next());
                    triple.setpCod(spIt._2.iterator().next());
                    triple.setoCod(tuple._2._2.iterator().next());
                    return triple;
                }

            });

            Dataset<Row> dfResult = sqlContext.createDataFrame(encodedRDD.rdd(), Triple.class);

            StructType schema = new StructType()
                    .add("sCod", "LONG", false)
                    .add("pCod", "LONG", false)
                    .add("oCod", "LONG", false)
                    .add("rn", "LONG", false);

            JavaRDD<Row> rdd = dfResult.select("sCod", "pCod", "oCod").orderBy(dfResult.col("sCod").asc(), dfResult.col("pCod").asc(), dfResult.col("oCod").asc())
                    .toJavaRDD().zipWithIndex().map(new Function<Tuple2<Row, Long>, Row>() {
                        @Override
                        public Row call(Tuple2<Row, Long> t1) throws Exception {
                            tripleStats.add(new Stats(1));
                            return RowFactory.create(t1._1.getLong(0), t1._1.getLong(1), t1._1.getLong(2), t1._2);
                        }
                    });

            dfResult = sqlContext.createDataFrame(rdd.rdd(), schema);
            if (outputParts > 0) {
                dfResult = dfResult.coalesce(outputParts).orderBy(dfResult.col("rn").asc());
            }
            dfResult.select("sCod", "pCod", "oCod").write().mode(SaveMode.Overwrite).format(outputFormat).save(outputPath + "triples.dat");

            datasetStats.put("tripleStats", tripleStats.value());
            long tdiff = System.currentTimeMillis() - t1;
            System.out.println(String.format("\t%d triples encoded in %.03f secs.", tripleStats.value().getCount(), (tdiff / 1000.0f)));

        } else if (joinMethod == JoinMethod.BROADCAST_MATERIALIZED) {
            System.out.println("Encoding triples with BROADCAST_MATERIALIZED...");

            long t1 = System.currentTimeMillis();

            Long soOffset = 1L;
            if (datasetStats.containsKey("soStats")) {
                soOffset += datasetStats.get("soStats").getCount();
            } else {
                soOffset += dict.get("so").count();
            }

            Dataset<Row> so = dict.get("so").withColumn("value2", dict.get("so").col("value").plus(1)).drop("value").withColumnRenamed("value2", "value");
            Dataset<Row> su = dict.get("su").withColumn("value2", dict.get("su").col("value").plus(soOffset)).drop("value").withColumnRenamed("value2", "value");
            Dataset<Row> pr = dict.get("pr").withColumn("value2", dict.get("pr").col("value").plus(1)).drop("value").withColumnRenamed("value2", "value");
            Dataset<Row> ob = dict.get("ob").withColumn("value2", dict.get("ob").col("value").plus(soOffset)).drop("value").withColumnRenamed("value2", "value");

            su = su.union(so);
            ob = ob.union(so);

            JavaPairRDD<String, Long> suRDD = su.javaRDD().mapToPair(new PairFunction<Row, String, Long>() {
                @Override
                public Tuple2<String, Long> call(Row t) throws Exception {
                    return new Tuple2<>(t.getString(0), t.getLong(1));
                }
            });

            JavaPairRDD<String, Long> prRDD = pr.javaRDD().mapToPair(new PairFunction<Row, String, Long>() {
                @Override
                public Tuple2<String, Long> call(Row t) throws Exception {
                    return new Tuple2<>(t.getString(0), t.getLong(1));
                }
            });

            JavaPairRDD<String, Long> obRDD = ob.javaRDD().mapToPair(new PairFunction<Row, String, Long>() {
                @Override
                public Tuple2<String, Long> call(Row t) throws Exception {
                    return new Tuple2<>(t.getString(0), t.getLong(1));
                }
            });
            final Broadcast<Map<String, Long>> smap = jsc.broadcast(suRDD.collectAsMap());
            final Broadcast<Map<String, Long>> pmap = jsc.broadcast(prRDD.collectAsMap());
            final Broadcast<Map<String, Long>> omap = jsc.broadcast(obRDD.collectAsMap());

            JavaRDD<Triple> encodedRDD = tripleRDD.map(
                    new Function<Triple, Triple>() {

                @Override
                public Triple call(Triple input) {
                    input.encodeSubject(smap.getValue());
                    input.encodePredicate(pmap.getValue());
                    input.encodeObject(omap.getValue());
                    return input;
                }
            });
            Dataset<Row> dfResult = sqlContext.createDataFrame(encodedRDD.rdd(), Triple.class);

            StructType schema = new StructType()
                    .add("sCod", "LONG", false)
                    .add("pCod", "LONG", false)
                    .add("oCod", "LONG", false)
                    .add("rn", "LONG", false);

            JavaRDD<Row> rdd = dfResult.select("sCod", "pCod", "oCod").orderBy(dfResult.col("sCod").asc(), dfResult.col("pCod").asc(), dfResult.col("oCod").asc())
                    .toJavaRDD().zipWithIndex().map(new Function<Tuple2<Row, Long>, Row>() {
                        @Override
                        public Row call(Tuple2<Row, Long> t1) throws Exception {
                            tripleStats.add(new Stats(1));
                            return RowFactory.create(t1._1.getLong(0), t1._1.getLong(1), t1._1.getLong(2), t1._2);
                        }
                    });

            dfResult = sqlContext.createDataFrame(rdd.rdd(), schema);

            if (outputParts > 0) {
                dfResult = dfResult.coalesce(outputParts).orderBy(dfResult.col("rn").asc());
            }
            dfResult.select("sCod", "pCod", "oCod").write().mode(SaveMode.Overwrite).format(outputFormat).save(outputPath + "triples.dat");

            smap.destroy();
            omap.destroy();
            pmap.destroy();

            datasetStats.put("tripleStats", tripleStats.value());
            long tdiff = System.currentTimeMillis() - t1;
            System.out.println(String.format("\t%d triples encoded in %.03f secs.", tripleStats.value().getCount(), (tdiff / 1000.0f)));
        }
    }

    /**
     *
     * @param outputParts
     * @param tripleRDD
     * @param doRepartition
     * @param numPartitions
     * @param doCache
     * @param doPersist
     * @param soStats
     * @param soKeyStats
     * @param subjStats
     * @param subjKeyStats
     * @param predStats
     * @param predKeyStats
     * @param objStats
     * @param objKeyStats
     * @param sqlContext
     * @param outputFormat
     * @param outputPath
     * @param datasetStats
     * @throws InterruptedException
     * @throws ExecutionException
     */
    private static void generateDictionary(final int outputParts, JavaRDD<Triple> tripleRDD, final boolean doRepartition, int numPartitions, final boolean doCache, final boolean doPersist, final StatsAccumulator soStats, final StatsAccumulator soKeyStats, final StatsAccumulator subjStats, final StatsAccumulator subjKeyStats, final StatsAccumulator predStats, final StatsAccumulator predKeyStats, final StatsAccumulator objStats, final StatsAccumulator objKeyStats, final SQLContext sqlContext, final String outputFormat, final String outputPath, Map<String, Stats> datasetStats) throws InterruptedException, ExecutionException {
        long t1 = System.currentTimeMillis();

        System.out.println("Generating dictionary...");

        System.out.println("\tWritting sorted dictionaries using dataframes.");

        final StructType triplesSchema = new StructType()
                .add("s", "STRING", true)
                .add("p", "STRING", true)
                .add("o", "STRING", true);

        JavaRDD<Row> triples = tripleRDD.map(new Function<Triple, Row>() {
            @Override
            public Row call(Triple t1) throws Exception {
                return RowFactory.create(t1.getSubject(), t1.getPredicate(), t1.getObject());
            }

        });
        Dataset<Row> tripleDF = sqlContext.createDataFrame(triples, triplesSchema);

        if (doPersist) {
            System.out.println("\tPersisting...");
        }

        Dataset<Row> su_p = tripleDF.select("s").withColumnRenamed("s", "key").distinct();
        if (doPersist) {
            su_p = su_p.persist();
        }
        Dataset<Row> pr = tripleDF.select("p").withColumnRenamed("p", "key").distinct();
        Dataset<Row> ob_p = tripleDF.select("o").withColumnRenamed("o", "key").distinct();
        if (doPersist) {
            ob_p = ob_p.persist();
        }

        Dataset<Row> so_p = su_p.intersect(ob_p);
        if (doPersist) {
            so_p = so_p.persist();
        }
        Dataset<Row> su = su_p.except(so_p);
        Dataset<Row> ob = ob_p.except(so_p);

        final StructType schemaOut = new StructType()
                .add("key", "STRING", true)
                .add("value", "LONG", true);

        Dataset<Row> df;

        System.out.println("\tWritting...");

        df = pr.sort(pr.col("key").asc());
        final Dataset<Row> df1 = sqlContext.createDataFrame(df.toJavaRDD().zipWithIndex().map(new Function<Tuple2<Row, Long>, Row>() {
            @Override
            public Row call(Tuple2<Row, Long> t1) throws Exception {
                predStats.add(new Stats(t1._2));
                predKeyStats.add(new Stats((long) t1._1.getString(0).length()));
                return RowFactory.create(t1._1.getString(0), t1._2);
            }
        }), schemaOut);

        df = su.sort(su.col("key").asc());
        final Dataset<Row> df2 = sqlContext.createDataFrame(df.toJavaRDD().zipWithIndex().map(new Function<Tuple2<Row, Long>, Row>() {
            @Override
            public Row call(Tuple2<Row, Long> t1) throws Exception {
                subjStats.add(new Stats(t1._2));
                subjKeyStats.add(new Stats((long) t1._1.getString(0).length()));
                return RowFactory.create(t1._1.getString(0), t1._2);
            }
        }), schemaOut);

        df = ob.sort(ob.col("key").asc());
        final Dataset<Row> df3 = sqlContext.createDataFrame(df.toJavaRDD().zipWithIndex().map(new Function<Tuple2<Row, Long>, Row>() {
            @Override
            public Row call(Tuple2<Row, Long> t1) throws Exception {
                objStats.add(new Stats(t1._2));
                objKeyStats.add(new Stats((long) t1._1.getString(0).length()));
                return RowFactory.create(t1._1.getString(0), t1._2);
            }
        }), schemaOut);

        df = so_p.sort(so_p.col("key").asc());
        final Dataset<Row> df4 = sqlContext.createDataFrame(df.toJavaRDD().zipWithIndex().map(new Function<Tuple2<Row, Long>, Row>() {
            @Override
            public Row call(Tuple2<Row, Long> t1) throws Exception {
                soStats.add(new Stats(t1._2));
                soKeyStats.add(new Stats((long) t1._1.getString(0).length()));
                return RowFactory.create(t1._1.getString(0), t1._2);
            }
        }), schemaOut);

        ExecutorService taskExecutor = Executors.newFixedThreadPool(4);
        List<Future> futures = new ArrayList<>();

        futures.add(taskExecutor.submit(new Callable<Integer>() {
            @Override
            public Integer call() throws Exception {
                Dataset<Row> dfout = df1;
                if (outputParts > 0) {
                    dfout = dfout.coalesce(outputParts).sort(dfout.col("value").asc());
                }
                dfout.write().mode(SaveMode.Overwrite).format(outputFormat).save(outputPath + "pr.dict");
                return 0;
            }
        }));
        futures.add(taskExecutor.submit(new Callable<Integer>() {
            @Override
            public Integer call() throws Exception {
                Dataset<Row> dfout = df2;
                if (outputParts > 0) {
                    dfout = dfout.coalesce(outputParts).sort(dfout.col("value").asc());
                }
                dfout.write().mode(SaveMode.Overwrite).format(outputFormat).save(outputPath + "su.dict");
                return 0;
            }
        }));
        futures.add(taskExecutor.submit(new Callable<Integer>() {
            @Override
            public Integer call() throws Exception {
                Dataset<Row> dfout = df3;
                if (outputParts > 0) {
                    dfout = dfout.coalesce(outputParts).sort(dfout.col("value").asc());
                }
                dfout.write().mode(SaveMode.Overwrite).format(outputFormat).save(outputPath + "ob.dict");
                return 0;
            }
        }));
        futures.add(taskExecutor.submit(new Callable<Integer>() {
            @Override
            public Integer call() throws Exception {
                Dataset<Row> dfout = df4;
                if (outputParts > 0) {
                    dfout = dfout.coalesce(outputParts).sort(dfout.col("value").asc());
                }
                dfout.write().mode(SaveMode.Overwrite).format(outputFormat).save(outputPath + "so.dict");
                return 0;
            }
        }));

        for (Future f : futures) {
            f.get();
        }
        System.out.println("\tDictionaries written");

        if (doPersist) {
            su_p.unpersist();
            ob_p.unpersist();
            so_p.unpersist();
        }

        datasetStats.put("predStats", predStats.value());
        datasetStats.put("predKeyStats", predKeyStats.value());
        datasetStats.put("subjStats", subjStats.value());
        datasetStats.put("subjKeyStats", subjKeyStats.value());
        datasetStats.put("objStats", objStats.value());
        datasetStats.put("objKeyStats", objKeyStats.value());
        datasetStats.put("soStats", soStats.value());
        datasetStats.put("soKeyStats", soKeyStats.value());

        long tdiff = System.currentTimeMillis() - t1;
        System.out.println(String.format("\tDict generated in %.03f secs.", (tdiff / 1000.0f)));
    }

    /**
     *
     * @param elements
     * @return
     */
    static JavaRDD<DictEntry> AsSortedKeyValueRDD(final JavaRDD<String> elements, final StatsAccumulator stats, final StatsAccumulator keyStats) {

        JavaPairRDD<String, Long> rdd = elements.mapToPair(
                new PairFunction<String, String, Long>() {
            @Override
            public Tuple2<String, Long> call(String t) throws Exception {
                return new Tuple2(t, 0L);
            }
        }
        ).reduceByKey(
                new Function2<Long, Long, Long>() {
            @Override
            public Long call(Long v1, Long v2) throws Exception {
                return 0L;
            }
        }).sortByKey(true);

        return rdd.keys().zipWithIndex().map(
                new Function<Tuple2<String, Long>, DictEntry>() {
            @Override
            public DictEntry call(Tuple2<String, Long> t) throws Exception {
                stats.add(new Stats(t._2));
                keyStats.add(new Stats((long) t._1.length()));
                return new DictEntry(t._1, t._2);
            }
        }
        );
    }

}
