package com.cbr.hdt.impl.spark;

import com.cbr.hdt.spark.Stats;
import java.io.IOException;
import java.net.URI;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.LocatedFileStatus;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.fs.RemoteIterator;
import org.apache.hadoop.io.SequenceFile;
import org.apache.parquet.column.page.PageReadStore;
import org.apache.parquet.example.data.Group;
import org.apache.parquet.example.data.simple.convert.GroupRecordConverter;
import org.apache.parquet.hadoop.ParquetFileReader;
import org.apache.parquet.hadoop.metadata.ParquetMetadata;
import org.apache.parquet.io.ColumnIOFactory;
import org.apache.parquet.io.MessageColumnIO;
import org.apache.parquet.io.RecordReader;
import org.apache.parquet.schema.MessageType;
import org.apache.spark.sql.Dataset;
import org.apache.spark.sql.Row;
import static org.apache.spark.sql.functions.count;
import static org.apache.spark.sql.functions.max;
import static org.apache.spark.sql.functions.min;
import org.rdfhdt.hdt.compact.bitmap.AdjacencyList;
import org.rdfhdt.hdt.enums.TripleComponentOrder;
import org.rdfhdt.hdt.exceptions.IllegalFormatException;
import org.rdfhdt.hdt.listener.ProgressListener;
import org.rdfhdt.hdt.options.HDTOptions;
import org.rdfhdt.hdt.triples.TripleID;
import org.rdfhdt.hdt.triples.impl.BitmapTriples;
import org.rdfhdt.hdt.triples.impl.TripleOrderConvert;
import org.rdfhdt.hdt.util.BitUtil;

/**
 *
 * @author Carlos Barrales Ruiz <cbruiz@gmail.com>
 */
public class SparkBitMapTriples extends BitmapTriples {

    public static int CHUNK_SIZE = 1 * 1024 * 1024;

    long number;
    long size;
    long lastX = 0, lastY = 0, lastZ = 0;
    long x, y, z;
    long numTriples = 0;
    boolean trimNeeded = false;

    Configuration conf = null;
    FileSystem fileSystem = null;
    Path path = null;

    Dataset<Row> df;

    /**
     *
     */
    public SparkBitMapTriples() {
        super();
    }

    /**
     *
     * @param spec
     * @param conf
     * @param pathString
     * @param df
     * @throws java.lang.Exception
     */
    public SparkBitMapTriples(HDTOptions spec, Configuration conf, String pathString, Dataset df) throws Exception {
        super(spec);
        this.path = new Path(pathString);
        this.conf = conf;
        this.conf.addResource("/etc/hadoop/conf/core-site.xml");
        this.conf.addResource("/etc/hadoop/conf/hdfs-site.xml");
        this.conf.set("fs.default.name", "hdfs://hadoop");
        this.fileSystem = FileSystem.get(new URI("hdfs://hadoop"), this.conf);
        this.df = df;
    }

    /**
     *
     * @param numentries
     * @param maxpredicate
     * @param maxobject
     * @throws IOException
     */
    public void initialize(long numentries, long maxpredicate, long maxobject) throws IOException {
        this.number = numentries;
        this.seqY = new TransientSequenceLog64(CHUNK_SIZE, BitUtil.log2(maxpredicate), this.number, null, null);
        this.seqZ = new TransientSequenceLog64(CHUNK_SIZE, BitUtil.log2(maxobject), this.number, null, null);
        this.bitmapY = new TransientBitmap375(CHUNK_SIZE, this.number, null, null);
        this.bitmapZ = new TransientBitmap375(CHUNK_SIZE, this.number, null, null);
    }

    /**
     *
     * @param input
     * @param listener
     * @throws IOException
     * @throws Exception
     */
    public void load(SequenceFile.Reader input, ProgressListener listener) throws IOException, Exception {
        throw new Exception("Not implemented");
    }

    /**
     *
     * @param listener
     * @param stats
     * @throws IOException
     */
    public void load(ProgressListener listener, Map<String, Stats> stats) throws IOException {

        long count;
        long minSubj;
        long maxSubj;
        long minPred;
        long maxPred;
        long minObj;
        long maxObj;
        
        if (stats.containsKey("tripleStats")
                && stats.containsKey("subjStats")
                && stats.containsKey("predStats")
                && stats.containsKey("objStats")
                && stats.containsKey("soStats")) {
                        
            count = stats.get("tripleStats").getCount();
            
            minSubj = stats.get("soStats").getCount() + 1;
            maxSubj = minSubj + stats.get("subjStats").getCount();
   
            minPred = stats.get("predStats").getMin() + 1;
            maxPred = minPred + stats.get("predStats").getCount();
            
            minObj = stats.get("soStats").getCount() + 1;
            maxObj = minObj + stats.get("objStats").getCount() ;
             
        }
        else {
            System.out.println("\t\tComputing triple stats...");
            Row dfStatsRow = df.agg(
                    count(df.col("sCod")),
                    min(df.col("sCod")),
                    max(df.col("sCod")),
                    min(df.col("pCod")),
                    max(df.col("pCod")),
                    min(df.col("oCod")),
                    max(df.col("oCod"))
            ).first();
            count = dfStatsRow.getLong(0);
            minSubj = dfStatsRow.getLong(1);
            maxSubj = dfStatsRow.getLong(2);
            minPred = dfStatsRow.getLong(3);
            maxPred = dfStatsRow.getLong(4);
            minObj = dfStatsRow.getLong(5);
            maxObj = dfStatsRow.getLong(6);
        }

        this.initialize(count, maxPred, maxObj);
        System.out.println(String.format("\t\tNum triples: %d\n\t\tSubject idx [ %d to %d ]\n\t\tPredicates idx [ %d to %d ]\n\t\tObjects idx [ %d to %d ]\n\t\tSubject/Objects idx [ %d to %d ]",
                count,
                minSubj, maxSubj,
                minPred, maxPred,
                minObj, maxObj,
                1, Math.max(1, Math.max(minSubj, minObj) - 1)
        ));
        TransientSequenceLog64 vectorY = (TransientSequenceLog64) this.seqY;
        TransientSequenceLog64 vectorZ = (TransientSequenceLog64) this.seqZ;
        TransientBitmap375 bitY = (TransientBitmap375) this.bitmapY;
        TransientBitmap375 bitZ = (TransientBitmap375) this.bitmapZ;
                        
        SortedMap<String, Path> sortedPaths = new TreeMap<>();

        System.out.println(String.format("\t\tReading %s", path.toUri().toString()));
        for (RemoteIterator<LocatedFileStatus> it = fileSystem.listFiles(path, false); it.hasNext(); ) {
            LocatedFileStatus fileStatus = it.next();
            Path filePath = fileStatus.getPath();
            String fileName = filePath.getName();
            if (!fileName.startsWith(".") && !fileName.startsWith("_") && !fileName.startsWith("tmp")) {
                sortedPaths.put(fileName, filePath);
            }
        }
        for (Path filePath : sortedPaths.values()) {
            //System.out.println(String.format("\t\tReading %s", filePath.toUri().toString()));
            ParquetMetadata readFooter = ParquetFileReader.readFooter(conf, filePath);
            MessageType schema = readFooter.getFileMetaData().getSchema();
            ParquetFileReader parquetFileReader = new ParquetFileReader(conf, filePath, readFooter.getBlocks(), schema.getColumns());
            PageReadStore pages = null;
            try {
                while (null != (pages = parquetFileReader.readNextRowGroup())) {
                    final long rows = pages.getRowCount();

                    final MessageColumnIO columnIO = new ColumnIOFactory().getColumnIO(schema);
                    final RecordReader<Group> recordReader = columnIO.getRecordReader(pages, new GroupRecordConverter(schema));
                    for (int i = 0; i < rows; i++) {
                        final Group row = recordReader.read();
                        this.x = (int) row.getLong(0, 0);
                        this.y = (int) row.getLong(1, 0);
                        this.z = (int) row.getLong(2, 0);

                        if (this.x == 0 || this.y == 0 || this.z == 0) {
                            throw new IllegalFormatException("None of the components of a triple can be null");
                        }

                        if (this.numTriples == 0) {
                            // First triple
                            vectorY.append(this.y);
                            vectorZ.append(this.z);
                        } else if (this.x != this.lastX) {
                            if (this.x != this.lastX + 1) {
                                System.out.println(String.format("Index: %d Previous: %d", x, lastX));
                                throw new IllegalFormatException("Upper level must be increasing and correlative.");
                            }
                            // X changed
                            bitY.append(true);
                            vectorY.append(this.y);

                            bitZ.append(true);
                            vectorZ.append(this.z);
                        } else if (this.y != this.lastY) {
                            if (this.y < this.lastY) {
                                throw new IllegalFormatException("Middle level must be increasing for each parent.");
                            }

                            // Y changed
                            bitY.append(false);
                            vectorY.append(this.y);

                            bitZ.append(true);
                            vectorZ.append(this.z);
                        } else if (this.z != this.lastZ) { // Añadido para quitar triples duplicados
                            if (this.z < this.lastZ) {
                                throw new IllegalFormatException("Lower level must be increasing for each parent.");
                            }

                            // Z changed
                            bitZ.append(false);
                            vectorZ.append(this.z);
                        }

                        this.lastX = this.x;
                        this.lastY = this.y;
                        this.lastZ = this.z;
                        this.numTriples++;
                    }
                }
            } finally {
                parquetFileReader.close();
            }
        }
        close();
    }

    /**
     *
     * @param triple
     */
    public void add(TripleID triple) {
        TransientSequenceLog64 vectorY = (TransientSequenceLog64) this.seqY;
        TransientSequenceLog64 vectorZ = (TransientSequenceLog64) this.seqZ;
        TransientBitmap375 bitY = (TransientBitmap375) this.bitmapY;
        TransientBitmap375 bitZ = (TransientBitmap375) this.bitmapZ;

        TripleOrderConvert.swapComponentOrder(triple, TripleComponentOrder.SPO, this.order);
        this.x = triple.getSubject();
        this.y = triple.getPredicate();
        this.z = triple.getObject();

        if (this.x == 0 || this.y == 0 || this.z == 0) {
            throw new IllegalFormatException("None of the components of a triple can be null");
        }

        if (this.numTriples == 0) {
            // First triple
            vectorY.append(this.y);
            vectorZ.append(this.z);
        } else if (this.x != this.lastX) {
            if (this.x != this.lastX + 1) {
                System.out.println(String.format("Index: %d Previous: %d", x, lastX));
                throw new IllegalFormatException("Upper level must be increasing and correlative.");
            }
            // X changed
            bitY.append(true);
            vectorY.append(this.y);

            bitZ.append(true);
            vectorZ.append(this.z);
        } else if (this.y != this.lastY) {
            if (this.y < this.lastY) {
                throw new IllegalFormatException("Middle level must be increasing for each parent.");
            }

            // Y changed
            bitY.append(false);
            vectorY.append(this.y);

            bitZ.append(true);
            vectorZ.append(this.z);
        } else if (this.z != this.lastZ) { // Añadido para quitar triples duplicados
            if (this.z < this.lastZ) {
                throw new IllegalFormatException("Lower level must be increasing for each parent.");
            }

            // Z changed
            bitZ.append(false);
            vectorZ.append(this.z);
        }

        this.lastX = this.x;
        this.lastY = this.y;
        this.lastZ = this.z;
        this.numTriples++;
    }

    /**
     *
     * @throws IOException
     */
    @Override
    public void close() throws IOException {
        TransientSequenceLog64 vectorY = (TransientSequenceLog64) this.seqY;
        TransientSequenceLog64 vectorZ = (TransientSequenceLog64) this.seqZ;
        TransientBitmap375 bitY = (TransientBitmap375) this.bitmapY;
        TransientBitmap375 bitZ = (TransientBitmap375) this.bitmapZ;

        bitY.append(true);
        bitZ.append(true);

        bitY.close();
        bitZ.close();

        vectorY.close();
        vectorZ.close();

        if (this.trimNeeded) {
            vectorY.aggresiveTrimToSize();
            vectorZ.trimToSize();
        }

        this.adjY = new AdjacencyList(this.seqY, this.bitmapY);
        this.adjZ = new AdjacencyList(this.seqZ, this.bitmapZ);
    }

    /**
     *
     */
    void cleanup() {
        //this.seqY.cleanup();

    }

    /**
     *
     * @return
     */
    @Override
    public long getNumberOfElements() {
        return this.number;
    }

    /**
     *
     * @return
     */
    @Override
    public long size() {
        return this.size;
    }

}
